/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd All Rights Reserved
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an AS IS BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <dirent.h>
#include <sys/stat.h>
#include <unistd.h>
#include <limits.h>
#include <strings.h>
#include <pthread.h>

#include <cstdlib>
#include <cstring>
#include <algorithm>
#include <unordered_map>
#include <vector>
#include <iterator>
#include <sstream>

#include "utils.h"
#include "log.h"

static int pfd[2];
static pthread_t loggingThread;

bool iCompare(const std::string& a, const std::string& b)
{
	return a.length() == b.length() &&
		std::equal(b.begin(), b.end(), a.begin(),
			[](unsigned char a, unsigned char b)
			{ return std::tolower(a) == std::tolower(b); });
}

bool iCompare(const std::string& a, int aOffset, const std::string& b, int bOffset, int length)
{
	return static_cast<int>(a.length()) - length >= aOffset &&
		static_cast<int>(b.length()) - length >= bOffset &&
		std::equal(b.begin() + bOffset, b.begin() + bOffset + length, a.begin() + aOffset,
			[](unsigned char a, unsigned char b)
			{ return std::tolower(a) == std::tolower(b); });
}

bool isManagedAssembly(const std::string& fileName)
{
	return iCompare(fileName, fileName.size()-4, ".dll", 0, 4) ||
		iCompare(fileName, fileName.size()-4, ".exe", 0, 4);
}

bool isNativeImage(const std::string& fileName)
{
	return iCompare(fileName, fileName.size()-7, ".ni", 0, 3);
}

std::string readSelfPath()
{
	char buff[PATH_MAX];
	ssize_t len = ::readlink("/proc/self/exe", buff, sizeof(buff)-1);
	if (len != -1) {
		buff[len] = '\0';
		return std::string(buff);
	}

	return "";
}

std::string concatPath(const std::string& path1, const std::string& path2)
{
	std::string path(path1);
	if (path.back() == PATH_SEPARATOR) {
		path.append(path2);
	} else {
		path += PATH_SEPARATOR;
		path.append(path2);
	}

	return path;
}

void appendPath(std::string& path1, const std::string& path2)
{
	if (path1.back() == PATH_SEPARATOR) {
		path1.append(path2);
	} else {
		path1 += PATH_SEPARATOR;
		path1.append(path2);
	}
}

std::string absolutePath(const std::string& path)
{
	std::string absPath;
	char realPath[PATH_MAX];
	if (realpath(path.c_str(), realPath) != nullptr && realPath[0] != '\0')
		absPath.assign(realPath);

	return absPath;
}

std::string baseName(const std::string& path)
{
	auto pos = path.find_last_of(PATH_SEPARATOR);
	if (pos != std::string::npos)
		return path.substr(0, pos);
	else
		return std::string(".");
	return path;
}

bool endWithIgnoreCase(const std::string& str1, const std::string& str2, std::string& fileName)
{
	std::string::size_type len1 = str1.length();
	std::string::size_type len2 = str2.length();
	if (len2 > len1)
		return false;

	int i = 0;
	bool result = std::all_of(str1.cend() - len2, str1.end(),
				[&i, &str2] (char x) {
					return std::tolower(x) == std::tolower(str2[i++]);
				});
	if (result)
		fileName = str1.substr(0, len1 - len2);

	return result;
}

bool fileNotExist(const std::string& path)
{
	struct stat sb;
	return stat(path.c_str(), &sb) != 0;
}

#ifdef NOT_USE_FUNCTION
static bool extCheckAndGetFileNameIfExist(const std::string& dir, const std::string& ext, struct dirent* entry, std::string& fileName)
{
	std::string fName(entry->d_name);
	if (fName.length() < ext.length() ||
			fHame.compare(fName.length() - ext.length(), ext.length(), ext) != 0) {
		return false;
	}

	std::string fullName = concatPath(dir, entry->d_name);
	switch (entry->d_type) {
		case DT_REG: break;
		case DT_LNK:
		case DT_UNKNOWN:
			if (fileNotExist(fullName))
				return false;
		default:
			return false;
	}

	fileName = fullName;

	return true;
}
#endif

std::string stripNiDLL(const std::string& path)
{
	std::string niPath(path);
	if (path.size() < 5) return niPath;
	if (!strncasecmp(path.c_str() + path.size() - 4, ".dll", 4))
		niPath = path.substr(0, path.size()-4);
	else if (!strncasecmp(path.c_str() + path.size() - 4, ".exe", 4))
		niPath = path.substr(0, path.size()-4);

	if (!strncasecmp(niPath.c_str() + niPath.size() - 3, ".ni", 3))
		return niPath.substr(0, niPath.size()-3);

	return niPath;
}

std::string joinStrings(const std::vector<std::string>& strings, const char* const delimeter)
{
	switch (strings.size()) {
		case 0:
			return "";
		case 1:
			return strings[0];
		default:
			std::ostringstream os;
			std::copy(strings.begin(), strings.end()-1, std::ostream_iterator<std::string>(os, delimeter));
			os << *strings.rbegin();
			return os.str();
	}
}

struct AssemblyFile {
	std::string noExt;
	std::string ext;
};

bool operator == (const AssemblyFile& lhs, const AssemblyFile& rhs)
{
	return lhs.noExt == rhs.noExt && lhs.ext == rhs.ext;
}

namespace std {
	template<>
	struct hash<AssemblyFile> {
		std::size_t operator () (const AssemblyFile& f) const {
			const std::size_t h1 = std::hash<std::string>{}(f.noExt);
			const std::size_t h2 = std::hash<std::string>{}(f.ext);

			return h1 ^ (h2 << 1);
		}
	};
}

void assembliesInDirectory(const std::vector<std::string>& directories, std::string& tpaList)
{
	std::map<std::string, std::string> assemblyList;
	std::map<std::string, std::string> tmpList;

	auto reader = [&assemblyList, &tmpList] (const char* path, const char* name) {
		std::string pathStr(path);
		if (isManagedAssembly(pathStr)) {
			std::string dllName = stripNiDLL(name);
			std::pair<std::map<std::string, std::string>::iterator, bool> ret;
			ret = tmpList.insert(std::pair<std::string, std::string>(dllName, pathStr));
			if (ret.second == false) {
				if (isNativeImage(pathStr))
					tmpList[dllName] = pathStr;
			}
		}
	};

	for (auto directory : directories) {
		scanFilesInDir(directory.c_str(), reader, 1);
		// merge scaned dll list to tpa list.
		// if the dll is already exist in the list, that is skipped.
		assemblyList.insert(tmpList.begin(), tmpList.end());
	}

	std::map<std::string, std::string>::iterator it;
	for (it = assemblyList.begin(); it != assemblyList.end(); it++)
		tpaList += it->second + ':';

	if (tpaList.back() == ':')
		tpaList.pop_back();
}

void scanFilesInDir(const char* directory, FileReader reader, unsigned int depth)
{
	DIR *dir;
	struct dirent* entry;
	bool isDir;

	dir = opendir(directory);

	if (dir == nullptr)
		return;

	std::vector<std::string> innerDirectories;

	while ((entry = readdir(dir)) != nullptr) {
		isDir = false;
		std::string path = concatPath(directory, entry->d_name);
		switch (entry->d_type) {
			case DT_REG: break;
			case DT_DIR:
				isDir = true;
				break;
			case DT_LNK:
			case DT_UNKNOWN:
				struct stat sb;
				if (stat(path.c_str(), &sb) == -1)
					continue;

				if (S_ISREG(sb.st_mode) || S_ISDIR(sb.st_mode))
					break;
			default:
				continue;
		}
		if (!isDir)
			reader(path.c_str(), entry->d_name);
		else if (depth > 1 && strcmp(entry->d_name, ".") && strcmp(entry->d_name, ".."))
			innerDirectories.push_back(path);
	}

	if (depth != 0)
		for (auto& d : innerDirectories)
			scanFilesInDir(d.c_str(), reader, depth-1);

	closedir(dir);
}

static void *stdlog(void*)
{
    ssize_t readSize;
    char buf[1024];

    while ((readSize = read(pfd[0], buf, sizeof buf - 1)) > 0) {
        if (buf[readSize - 1] == '\n') {
            --readSize;
        }

        buf[readSize] = 0;

        _ERRX("%s", buf);
    }

    return 0;
}

int runLoggingThread() { // run this function to redirect your output to android log
    if (setvbuf(stdout, NULL, _IOLBF, 0) < 0) {
		_DBG("fail to make stdout line-buffered");
		return -1;
    }

    if (setvbuf(stderr, NULL, _IONBF, 0) < 0) {
		_DBG("make stderr unbuffered");
		return -1;
	}

    /* create the pipe and redirect stdout and stderr */
    if (pipe(pfd) < 0) {
		_DBG("fail to create pipe for logging");
		return -1;
    }

    if (dup2(pfd[1], fileno(stdout)) == -1) {
		_DBG("fail to duplicate fd to stdout");
		return -1;
    }

    if (dup2(pfd[1], fileno(stderr)) == -1) {
		_DBG("fail to duplicate fd to stderr");
		return -1;
	}

    /* spawn the logging thread */
    if (pthread_create(&loggingThread, 0, stdlog, 0) != 0) {
		_DBG("fail to create pthread");
        return -1;
    }

    if (pthread_detach(loggingThread) != 0) {
		_DBG("fail to detach pthread");
		return -1;
	}

    return 0;
}

